# This configuration script is compatible with MPC564xB-C devices.
# Features:
# - Calls system reset.
# -    Autodetects the current MPC564xB/C part
# - Disables the Software Watchdog Timer (SWT)
# - Initializes the MMU for the z4 core, single 4GB addrss space
# - Initializes the system ECC SRAM 
# - Sets the PC to the user program entry point from the internal flash
# If the entry point isn't available the BAM start address.
#
# Rev. 1.2 - fix reading RCHW 
# Rev. 1.3 - set IVORx
#
# VERSION: 1.3
# Obs: This version initializes and uses only the z4 core and it is compatible
#         with CodeWarrior for MCU 10.2.x.

######################################
# Initialize target variables
######################################
# var booke_vle: BOOKE = booke, VLE = vle
# only for 
set booke_vle vle

# SRAM memory sizes for supported devices.
# $mem_size($current_cfg) gives the current RAM to init
# CFG = xy from MPC564xy
# CFG         Mem Size
# ------------------------
# 4B        128K (0x20000)
# 4C         192K (0x30000)
# 5B        160K (0x28000)
# 5C         256K (0x40000)
# 6B        192K (0x30000)
# 6C         256K (0x40000)
# ------------------------
set mem_size(5644B) 0x20000
set mem_size(5644C) 0x30000
set mem_size(5645B) 0x28000
set mem_size(5645C) 0x40000
set mem_size(5646B) 0x30000
set mem_size(5646C) 0x40000
# The current configuration will be updated by detect_mpc564xbc_device
set current_cfg 5646C

# GPR registrer group
set GPR_GROUP "General Purpose Registers/"
# special purpose register group
set SPR_GROUP "e200z4 Special Purpose Registers/"
#TLB1 registers group
set TLB1_GROUP "regPPCTLB1/"

# SSCM register address
set SSCM_STATUS_ADDR 0xC3FD8000

# RCHW addresses of each Flash boot sector
#array set FLASH_BOOT_RCHW {0 0x0 1 0x4000 2 0x8000 3 0x10000 4 0x18000 5 0x1C000 6 0x20000 7 0x30000}
set SECTORS_ADD [list 0x0 0x8000 0xC000 0x10000 0x18000]

#################################################
# Initialize a RAM 'range' from 'start' address,
# downloading the init program at 0x4000_0000.
#################################################
proc init_ram {start range} {
      global GPR_GROUP
      global booke_vle

    puts "init ECC SRAM $start:$range"
      # default SRAM address
      set pstart 0x40000000
      # the offset from the given start address
      # at which the init prgram starts to init SRAM
    set offset 0x0
    # stmw write page size = 128 = 4bytes * 32 GPRS
    set psize 0x80

    if {$start == $pstart} {
        # init first 4 bytes (mem access) x 128 = 512
        # bytes to avoid reading the memory around PC
        # after stopping the core
        mem $start 256 = 0x0
        # base init address
        set offset 0x80
    }
    
    # address to start initialization
    set start [expr {$start + $offset}]
    
    # load add into GPR
    reg ${GPR_GROUP}GPR11 %d = $start

    # compute actual times stmw is called
    # and set counter
    set c [expr ((($range - $offset)/$psize))]
    reg ${GPR_GROUP}GPR12 %d = $c

    # execute init ram code
    if {$booke_vle == "vle"} {
        #mtctr r12
        mem $pstart = 0x7D8903A6
        #stmw r0,0(r11)
        mem [format %x [expr $pstart + 0x4]] = 0x180B0900
        #addi r11,r11,128
        mem [format %x [expr $pstart + 0x8]] = 0x1D6B0080
        #bdnz -8
        mem [format %x [expr $pstart + 0xc]] = 0x7A20FFF8
        # infinte loop
        #se_b *+0
        mem [format %x [expr $pstart + 0x10]] = 0xE8000000
    } else {
        #mtctr r12
        mem $pstart = 0x7D8903A6
        #stmw r0,0(r11)
        mem [format %x [expr $pstart + 0x4]] = 0xBC0B0000
        #addi r11,r11,128
        mem [format %x [expr $pstart + 0x8]] = 0x396B0080
        #bdnz -8
        mem [format %x [expr $pstart + 0xc]] = 0x4200FFF8
        # infinte loop
        #se_b *+0
        mem [format %x [expr $pstart + 0x10]] = 0x48000000    
    }
    
    # set PC to the first init instruction
    reg ${GPR_GROUP}PC = $pstart
    # execute init ram code
    # timeout 1 second to allow the code to execute
    go 1
    stop
}

#########################################
## Detects the current MPC564xB-C device
#########################################
proc detect_mpc564xbc_device {} {
    global current_cfg
    set SIUL_MIDR1 0x0
    set SIUL_MIDR2 0x0
    
    puts "reading SIUL.MIDR1 ..."
    set r [catch {set SIUL_MIDR1 [mem 0xC3F90004]}]
    if {$r == 0} {
        # check if it
        set r [expr ($SIUL_MIDR1 & 0xFFF00000) == 0x56400000]
        if {$r == 1} {
            set partnum1 [expr ($SIUL_MIDR1 & 0xFFFF0000) >> 16]
            puts "reading SIUL.MIDR2 ..."
            set SIUL_MIDR2 [mem 0xC3F90008]
            set partnum2 [expr ($SIUL_MIDR2 & 0x0000FF00) >> 8]
            if {$partnum2 == 66} {
                set current_cfg [format "%xB" $partnum1]
            } elseif {$partnum2 == 67} {
                set current_cfg [format "%xC" $partnum1]
            } else {
                puts "Wrong partnum in MIDR2."
            }
            puts [format "Current device: MPC%s, mask %x" $current_cfg [expr $SIUL_MIDR2 & 0x000000FF]]
        } else {
            puts [format "Incompatible device: MPC%x. Use default device MPC%s" [expr ($SIUL_MIDR1 & 0xFFFF0000) >> 16] $current_cfg]
        }
    } else {
        puts "Unable to detect the device."
    }
}

#################################################################
# Tries to obtain the program entry point from binary debug info
# or by searching for the first flash boot sector. If neither
# is succesfull it returns the BAM start address.
#################################################################
proc get_entry_point {} {
    global SECTORS_ADD
    # check symbolic names first
    set reset_vector_addr 0x3
    catch {set reset_vector_addr [evaluate __startup]}

    # If no debug info available search for a FLASH boot sector
    if {$reset_vector_addr == 0x3} {
        puts "no syms available. searching for boot sectors ..."
        set rchw_value 0x0
        foreach rchw $SECTORS_ADD {
            catch {set rchw_value [mem $rchw -np]}
            if {[expr $rchw_value & 0xFF0000] == 0x5A0000} {
                catch {set reset_vector_addr [mem [format "0x%x" [expr $rchw + 0x4]] -np]}
                set reset_vector_addr [format "0x%x" [expr $reset_vector_addr & 0xFFFFFFFF]]
                puts "found boot sector at $rchw and entry point at $reset_vector_addr."  
                break
            }
        }        
    }
    if {$reset_vector_addr == 0x3} {
        puts "no entry point available. return BAM start address."
        set reset_vector_addr 0xFFFFC000
    }
    return $reset_vector_addr
}


proc mpc564xbc_init {} {
    global SPR_GROUP
    global GPR_GROUP
    global TLB1_GROUP
    global booke_vle
    global current_cfg
    global mem_size
    
    reset hard
    
    # Explicitly stop Core0
    stop
    
    #initialize Z4 MMU
    # Setup MMU entry 15 for entire 32-bit address space: base address = 0x0000_0000
    # 4GB Memory Space, Not Guarded, Cache inhibited, All Access
    if {$booke_vle == "vle"} {
        # VLE page
        reg ${TLB1_GROUP}MMU_CAM0 = 0xB0000008FE0800000000000000000001
    } else {
        # BOOKE page
        reg ${TLB1_GROUP}MMU_CAM0 = 0xB0000008FC0800000000000000000001
    }

    # check single chip mode, SSCM.STATUS.BMODE == 0b011
    set SSCM_STATUS 0x0
    puts "reading SSCM.STATUS ..."
    catch {set SSCM_STATUS [mem 0xC3FD8000]}
    set r [expr ($SSCM_STATUS & 0xE00000)]

    # Besides single chip mode the script shouldn't do any
    # initialization, eg. ECC SRAM or set the entry point.
    if {$r == 0x600000} {
        puts "processsor in single chip mode. start initialization ..."
        
        # ensure SWT is disabled
        mem 0xFFF38010 = 0x0000C520
        mem 0xFFF38010 = 0x0000D928
        mem 0xFFF38000 = 0x8000010A
        
        # reset watchdog timer
        change ${SPR_GROUP}TCR = 0x0
        
        # detect MPC564xBC device
        detect_mpc564xbc_device
    
        #initialze ECC SRAM    
        init_ram 0x40000000 $mem_size($current_cfg)
    
        # initialize PC after reset
        reg ${GPR_GROUP}PC = [get_entry_point]
        
        # set SP to an unaligned (4bytes) to avoid creating invalid stack frames
        reg ${GPR_GROUP}SP = 0x3

        # Initialize core interrupts in order to catch any exceptions during
        # MCU initialization. The following sequence queries symbolics to
        # obtain the address of the wellknown symbol "ivor_branch_table_p0".
        set ivor_table 0x40000000
        set r [catch {set ivor_table [evaluate ivor_branch_table_p0]}]
        if {$r == 0} {
            # set IVPR vector base table high 16-bit and IVOR0 lower 16-bit
            set ivor0 0x[format %x [expr $ivor_table & 0xFFFF]]
            reg ${SPR_GROUP}IVPR = 0x[format %x [expr $ivor_table & 0xFFFF0000]]
            # Initiliaze only the first 16 IVORs
            for {set i 0} {$i < 16} {incr i} {
                # The IVOR branch table must have a 16 byte aligment.
                reg ${SPR_GROUP}IVOR${i} = 0x[format %x [expr $ivor0 + $i * 0x10]]
            }
        }
    } else {
        puts "processsor not in single chip mode. initialization not required."
        # set BAM address instead
        reg ${GPR_GROUP}PC = 0xFFFFC000
        # set SP to an unaligned (4bytes) to avoid creating invalid stack frames
        reg ${GPR_GROUP}SP = 0x3
    }
}

proc envsetup {} {
    # Environment Setup
    radix x 
    config hexprefix 0x
    config MemIdentifier v 
    config MemWidth 32 
    config MemAccess 32 
    config MemSwap off
}

#-------------------------------------------------------------------------------
# Main                                                                          
#-------------------------------------------------------------------------------
       
  envsetup

  mpc564xbc_init
